(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import predicate.
Require Import closure.
Require Import carrier. 
Require Import finite.

Require Import labels.
Require Import formula.
Require Import words.
Require Import constraints.
Require Import pmo.
Require Import css.

Require Import branch.
Require Import bi_expansion.

Inductive bi_closed_branch_cond : CSS_type -> Prop :=
  | in_bi_cb_0 : forall (X : _->Prop) C A m n, X (T A # m) -> X (F A # n) -> pmo C (m ~~ n) -> bi_closed_branch_cond (X,C)
  | in_bi_cb_1 : forall (X : _->Prop) C m, X (F I # m) -> pmo C (eps~~m) -> bi_closed_branch_cond (X,C)
  | in_bi_cb_2 : forall (X : _->Prop) C m, X (T BB # m) -> bi_closed_branch_cond (X,C).

Definition bi_closed_branch := CSS cap bi_closed_branch_cond.

Hint Extern 2 ( bi_closed_branch _ ) => apply bin_inter_sg'. 

Section bi_cb_inc.

  Let bi_cb_inc_cond c1 c2 : fst c1 inc fst c2 -> pmo (snd c1) inc pmo (snd c2) -> bi_closed_branch_cond c1 -> bi_closed_branch_cond c2.
  Proof.
    intros HX HC H.  
    induction H as [ X C A m n H1 H2 H3 | X C m H1 H2 | X C m H1 ]; destruct c2 as (X',C'); simpl in HX, HC.
    
    apply HX in H1.
    apply HX in H2.
    apply HC in H3.
    constructor 1 with (1 := H1) (3 := H3); auto.
    
    apply HX in H1.
    apply HC in H2.
    constructor 2 with (1 := H1); auto.
    
    apply HX in H1.
    constructor 3 with (1 := H1); auto.
  Qed.
      
  Proposition bi_cb_inc X C X' C' : X inc X' -> pmo C inc pmo C' -> CSS (X',C') -> bi_closed_branch (X,C) -> bi_closed_branch (X',C').
  Proof.
    intros H1 H2 H3 H4.
    apply bin_inter_sg in H4.
    destruct H4 as [ H4 H5 ].
    apply bin_inter_sg'; auto.
    apply bi_cb_inc_cond with (3 := H5); auto.
  Qed.

End bi_cb_inc.

Proposition bi_cb_inc_carrier : forall l1 l2, listbranch_eq CSS_inc l1 l2 -> carrier l1 inc bi_closed_branch -> carrier l2 inc bi_closed_branch.
Proof.
  apply listbranch_eq_lift.
  intros [] [] ( H1 & H2 & H3 ) H.
  generalize H.
  apply bi_cb_inc ; auto.
  apply H1.
  apply bin_inter_sg in H.
  intros s f m Hf. 
  apply H1 in Hf.  
  apply H2, (proj1 H) with (1 := Hf).
Qed.
  
Proposition bi_cb_subst sigma css : bi_closed_branch css -> bi_closed_branch (subst_CSS sigma css).
Proof.
  intros H.
  apply bin_inter_sg in H.
  destruct H as [ Hcss H2 ].
  apply bin_inter_sg'.
  destruct css; apply CSS_subst_stable; auto.
  induction H2 as [ X C A m n H1 H2 H3 | X C m H1 H2 | X C m H1 ].
  
  apply subst_pmo_prop with (sigma := sigma) in H3.
  constructor 1 with (A := A) (3 := H3).
  exists (T A # m); auto.
  exists (F A # n); auto.
  
  apply subst_pmo_prop with (sigma := sigma) in H2.
  simpl in H2.
  constructor 2 with (2 := H2).
  exists (F I # m); auto.
  
  simpl.
  constructor 3 with (m := subst_w sigma m).
  exists (T BB # m); auto.
Qed.

Hint Resolve bi_cb_subst.

Proposition bi_cb_subst_carrier sigma : forall lcss, carrier lcss                         inc bi_closed_branch 
                                                  -> carrier (map (subst_CSS sigma) lcss) inc bi_closed_branch.
Proof. apply list_lift; auto. Qed.

Proposition bi_cb_expand : forall css1 css2, CSS css1 -> bi_closed_branch css2 -> bi_closed_branch (expand_1 css1 css2).
Proof.
   intros (X1,C1) (X,C) H1 HXC.
   apply bi_cb_inc with (4 := HXC); auto.
   apply bin_inter_sg in HXC; destruct HXC; auto.
Qed.

Hint Resolve bi_cb_expand.

Proposition bi_cb_expand_carrier css : CSS css -> forall lcss, carrier lcss inc bi_closed_branch -> carrier (expand css lcss) inc bi_closed_branch.
Proof. intro; apply list_lift; auto. Qed.

Proposition bi_cb_finite css : bi_closed_branch css -> exists cssf, bi_closed_branch cssf /\ CSS_leqf cssf css.
Proof.
  intros H.
  apply bin_inter_sg in H.
  destruct H as [ Hcss H2 ].
  induction H2 as [ X C A m n H1 H2 H3 | X C m H1 H2 | X C m H1 ].
  destruct (pmo_cpct H3) as ( Cf & H4 & H5 & H6 ).
  assert (pmo Cf (m ~~ m)) as Hm.
    apply ceq_lft with (2 := H6); auto.
  assert (pmo Cf (n ~~ n)) as Hn.
    apply ceq_rt  with (2 := H6); auto.
  exists (sg (T A # m) cup sg (F A # n), Cf).
  split; auto.
  apply bin_inter_sg'.
  intros s f x Hx; apply bin_union_sg in Hx.
  destruct Hx as [ Hx | Hx ]; apply in_sg_eq in Hx; injection Hx; clear Hx; intros Hx _ _ ; rewrite <- Hx; auto.
  constructor 1 with (A := A) (m := m) (n := n); apply in_sg; auto.

  destruct (pmo_cpct H2) as ( Cf & H4 & H5 & H6 ).
  assert (pmo Cf (m ~~ m)) as Hm.
    apply ceq_rt with (2 := H6); auto.
  exists (sg (F I # m), Cf).
  split; auto.
  apply bin_inter_sg'.
  intros s f x Hx; apply in_sg_eq in Hx; injection Hx; clear Hx; intros Hx _ _ ; rewrite <- Hx; auto.
  constructor 2 with (m := m); apply in_sg; auto.
  
  generalize (Hcss _ _ _ H1); intros H2.
  destruct (pmo_cpct H2) as ( Cf & H4 & H5 & H6 ).
  assert (pmo Cf (m ~~ m)) as Hm.
    apply ceq_rt with (2 := H6); auto.
  exists (sg (T BB # m), Cf).
  split; auto.
  apply bin_inter_sg'.
  intros s f x Hx; apply in_sg_eq in Hx; injection Hx; clear Hx; intros Hx _ _ ; rewrite <- Hx; auto.
  constructor 3 with (m := m); apply in_sg; auto.
Qed.


  