(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import predicate.
Require Import closure.

Require Import labels.
Require Import words.
Require Import constraints.
Require Import pmo.

Require Import formula.
Require Import css.

Require Import kripke.

Section Hintikka.

  Variable X : stm -> Prop.
  Variable C : cst -> Prop.

  Definition htk_0  := CSS (X,C).
  Definition htk_1  := forall A m n, X (T A # m) -> X (F A # n) -> pmo C (m~~n) -> False.
  Definition htk_2  := forall A B m, X (T A %% B # m) -> X (T A # m) \/ X(T B # m).
  Definition htk_3  := forall A B m, X (F A %% B # m) -> X (F A # m) /\ X(F B # m).
  Definition htk_2' := forall A B m, X (T A #> B # m) -> forall x, pmo C (m ~~ x) -> X (F A # x) \/ X(T B # x).
  Definition htk_3' := forall A B m, X (F A #> B # m) -> exists x, X (T A # x) /\ X(F B # x) /\ pmo C (m ~~ x).
  Definition htk_4  := forall A B m, X (T A && B # m) -> X (T A # m) /\ X(T B # m).
  Definition htk_5  := forall A B m, X (F A && B # m) -> X (F A # m) \/ X(F B # m).
  Definition htk_6  := forall m, X (T I # m) -> pmo C (eps~~m).
  Definition htk_7  := forall m, X (F I # m) -> pmo C (eps~~m) -> False.
  Definition htk_6' := forall m, X (T BB # m) -> False.
  Definition htk_8  := forall A B m, X (T A * B # m)  -> exists x, exists y, X (T A # x) /\ X (T B # y) /\ pmo C (x o y ~~ m).
  Definition htk_9  := forall A B m, X (F A * B # m)  -> forall x y, pmo C (x o y ~~ m) -> X (F A # x) \/ X (F B # y).
  Definition htk_10 := forall A B m, X (T A -* B # m) -> forall x y, pmo C (x o m ~~ y) -> X (F A # x) \/ X (T B # y).
  Definition htk_11 := forall A B m, X (F A -* B # m) -> exists x, exists y, X (T A # x) /\ X (F B # y) /\ pmo C (x o m ~~ y).

  Inductive Hintikka :=
    | In_Hintikka : htk_0 -> htk_1 -> htk_2 -> htk_3 
                 -> htk_2'-> htk_3' 
                 -> htk_4 -> htk_5 
                 -> htk_6 -> htk_7 
                 -> htk_6'
                 -> htk_8 -> htk_9 -> htk_10 -> htk_11 
                 -> Hintikka.

  Definition bi_herbrand m v := exists n, pmo C (n~~m) /\ X (T (%v) # n).

  Proposition bi_herbrand_prop : bi_frame (pmo C) bi_herbrand.
  Proof.
    split.
    apply pmo_is_pmo.
    intros v m n Hmn ( k & H1 & H2 ).
    exists k.
    split; auto.
    apply pmo_tran with (1 := H1); auto.
  Defined.

  Notation "m '|=' f" := (forces (pmo C) bi_herbrand m f).

  Let mdl_stm_f f := forall m, (X (T f # m) ->   m |= f) 
                            /\ (X (F f # m) -> ~ m |= f).

  Section props.

    Hypothesis H0  : htk_0.
    Hypothesis H1  : htk_1.
    Hypothesis H2  : htk_2.
    Hypothesis H3  : htk_3.
    Hypothesis H2' : htk_2'.
    Hypothesis H3' : htk_3'.
    Hypothesis H4  : htk_4.
    Hypothesis H5  : htk_5.
    Hypothesis H6  : htk_6.
    Hypothesis H7  : htk_7.
    Hypothesis H6' : htk_6'.
    Hypothesis H8  : htk_8.
    Hypothesis H9  : htk_9.
    Hypothesis H10 : htk_10.
    Hypothesis H11 : htk_11.

    Variable f1 f2 : formula.

    Hypothesis Hf1 : mdl_stm_f f1.
    Hypothesis Hf2 : mdl_stm_f f2.

    Proposition Hintikka_or: mdl_stm_f (f1 %% f2).
    Proof.
      intros m; split; intros HX.
    
      generalize (H2 _ _ _ HX); intros [ H | H ]; 
      [ apply Hf1 in H; left | apply Hf2 in H; right ]; auto.
    
      simpl; intros [ H | H ]; generalize (H3 _ _ _ HX);
      intros [ G1 G2 ]; [ apply Hf1 in G1 | apply Hf2 in G2 ]; auto.
    Qed.

    Proposition Hintikka_imp: mdl_stm_f (f1 #> f2).
    Proof.
      intros m; split; intros HX.
      
      simpl; intros x Hx1 Hx2.
      generalize (H2' _ _ _ HX _ Hx1); intros [ H | H ]; 
      [ apply Hf1 in H | apply Hf2 in H ]; tauto.
    
      simpl; intros H. 
      destruct (H3' _ _ _ HX) as ( x & G1 & G2 & G3 ).
      specialize H with (1:= G3).
      apply Hf1 in G1.
      apply Hf2 in G2.
      auto.
    Qed.

    Proposition Hintikka_and: mdl_stm_f (f1 && f2).
    Proof.
      split; intros HX.

      generalize (H4 _ _ _ HX); intros (H,H'). 
      split; [ apply Hf1 | apply Hf2 ]; auto.
    
      simpl; intros (H,H'); generalize (H5 _ _ _ HX).
      intros [ G | G ]; [ apply Hf1 in G | apply Hf2 in G ]; apply G; auto.
    Qed.

    Proposition Hintikka_BB : mdl_stm_f BB.
    Proof.
      split.
      intros H; apply H6' in H; auto.
      auto.
    Qed.

    Proposition Hintikka_unit : mdl_stm_f I.
    Proof.
      intros m; split; intros HX.
      apply H6 in HX; auto.
      apply H7 in HX.
      intros H; apply HX; auto.
    Qed.

    Proposition Hintikka_mult: mdl_stm_f (f1 * f2).
    Proof.
      intros m; split; intros HX.

      generalize (H8 _ _ _ HX); intros ( x & y & H & H' & H'' ).
      apply Hf1 in H. apply Hf2 in H'.
      exists x; exists y; split; auto.
  
      simpl; intros ( x & y & H & H' & H'' ).
      generalize (H9 _ _ _ HX); intros G. 
      apply G in H.
      destruct H as [ H | H ]; [ apply Hf1 in H | apply Hf2 in H ]; apply H; auto.
    Qed.

    Proposition Hintikka_wand: mdl_stm_f (f1 -* f2).
    Proof.
      intros m; split; intros HX.

      simpl; intros x y H H'.
      generalize (H10 _ _ _ HX); intros G. 
      apply G in H.
      destruct H as [ H | H ]; [ apply Hf1 in H | apply Hf2 in H ]; tauto.

      generalize (H11 _ _ _ HX); intros ( x & y & H & H' & H'' ).
      apply Hf1 in H. apply Hf2 in H'.
      simpl. 
      intros G; apply G in H''; auto.
    Qed.

  End props.

  Hypothesis Htk : Hintikka.

  Let model_stm f : mdl_stm_f f. 
  Proof.
    destruct Htk as [ H0 H1 H2 H3 H2' H3' H4 H5 H6 H7 H8 H9 H10 H11 ].
    induction f as [ v | [] f IHf | [] f1 IHf1 f2 IHf2 ]; intros m.
 
    simpl; split.
    intros H. exists m. split; auto. apply H0 with (1 := H). 
    intros H ( n & G1 & G2 ). apply (H1 _ _ _ G2 H); auto.

    apply Hintikka_unit; auto.
    apply Hintikka_BB; auto.
    apply Hintikka_mult; auto.
    apply Hintikka_wand; auto.
    apply Hintikka_and; auto.
    apply Hintikka_or; auto.
    apply Hintikka_imp; auto.
  Qed.

  Proposition Hintikka_model : mdl_CSS (pmo C) bi_herbrand (fun a => l a) (X,C).
  Proof.
    destruct Htk as [ H0 ].
    split. apply H0.
    split. 
    intros ([],(f,m)) HX; simpl; rewrite subst_w_id;
    [  apply (proj1 (model_stm _ _) HX) 
    |  apply (proj2 (model_stm _ _) HX) ].
    intros (m,n) H; simpl; do 2 rewrite subst_w_id; apply pmo_inc; auto.
  Qed.

End Hintikka.
    

