(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import list_fall_exst.
Require Import predicate.
Require Import closure.
Require Import maps.
Require Import carrier. 
Require Import finite.

Require Import labels.
Require Import formula.
Require Import words.
Require Import substitutions.
Require Import constraints.
Require Import css.
Require Import pmo.

Require Import branch.
Require Import tableaux.

Require Import bi_expansion.
Require Import bi_exp_props.
Require Import bi_closed_branch.

Definition bi_tab := tableau CSS_id bi_expansion.

Definition bi_has_closed_tableau := has_closed_tableau CSS_id bi_expansion bi_closed_branch.

Definition bi_has_no_closed_tableau css := finite_CSS css /\ ~ has_closed_tableau CSS_id bi_expansion bi_closed_branch css.

Proposition bi_tab_CSS css t : CSS css -> bi_tab css t -> carrier t inc CSS.
Proof. intros Hcss ?; apply (tableau_prop_conj _ _ bi_expansion _ _ Hcss CSS_id_id bi_expansion_CSS); auto. Qed.
 
Proposition bi_tab_finite_CSS css t : finite_CSS css -> bi_tab css t -> carrier t inc finite_CSS.
Proof. intros Hcss ?; apply (tableau_prop_conj _ _ bi_expansion _ _ Hcss fCSS_id bi_expansion_finite_CSS); auto. Qed.

Proposition bi_tab_increase css t : bi_tab css t -> forall cssi, carrier t cssi -> CSS_leq css cssi.
Proof.
  apply tableau_prop_conj; auto.
  
  intros x y H0 H1.
  apply CSS_leq_trans with (2 := CSS_id_imp_leq _ _ H0); auto.
  
  intros b lb H0 H1 x Hx.
  apply CSS_leq_trans with (1 := H0).  
  apply bi_expansion_increase' with (1 := H1); auto.
Qed.

Proposition bi_tab_CSS_id css1 css2 t : CSS_id css1 css2 -> bi_tab css1 t -> bi_tab css2 t.
Proof.
  intro; apply tableau_eq; auto.
  apply CSS_id_trans.
Qed.

Proposition bi_tab_comp b0 l1 l2 b t :  bi_tab b t 
                                     -> bi_tab b0 (l1 ++ b :: l2)
                                     -> bi_tab b0 (l1 ++ t ++ l2).
Proof. 
  apply tableau_comp; auto.
  apply CSS_id_trans.
Qed.

Section BI_Tableaux_basic.

  Let bi_tableau_exp css lcss : bi_expansion css lcss -> bi_tab css lcss.
  Proof.
    intros H.
    cutrewrite (lcss = nil++lcss++nil).
    apply tableau_next with (2 := H); auto.
    apply tableau_start; auto.
    rewrite <- app_nil_end; auto.
  Qed.

  Variable X : stm -> Prop.
  Variable C : cst -> Prop.

  Hint Resolve bi_exp_Tand bi_exp_Fand bi_exp_Tor bi_exp_For bi_exp_Timp bi_exp_Fimp.
  Hint Resolve bi_exp_TI  bi_exp_Tmult  bi_exp_Fmult  bi_exp_Twand  bi_exp_Fwand.

  Ltac solve := intros; apply bi_tableau_exp; eauto.
 
  Proposition bi_tableau_exp_Tand A B m : X (T A && B # m) -> bi_tab (X,C) ((X cup (sg (T A # m) cup sg (T B # m)), C cup empty)::nil).
  Proof. solve. Qed.
  
  Proposition bi_tableau_exp_Fand A B m : X (F A && B # m) -> bi_tab (X,C) ((X cup sg (F A # m), C cup empty)::
                                                                            (X cup sg (F B # m), C cup empty)::nil).
  Proof. solve. Qed.

  Proposition bi_tableau_exp_Tor A B m : X (T A %% B # m) -> bi_tab (X,C) ((X cup sg (T A # m), C cup empty)::
                                                                           (X cup sg (T B # m), C cup empty)::nil).
  Proof. solve. Qed.

  Proposition bi_tableau_exp_For A B m : X (F A %% B # m) -> bi_tab (X,C) ((X cup (sg (F A # m) cup sg (F B # m)), C cup empty)::nil).
  Proof. solve. Qed.
  
  Proposition bi_tableau_exp_Timp A B m x : X (T A #> B # m) -> pmo C (m ~~ x) -> bi_tab (X,C) ((X cup sg (F A # x), C cup empty)::
                                                                                                (X cup sg (T B # x), C cup empty)::nil).
  Proof. solve. Qed.

  Proposition bi_tableau_exp_Fimp A B m a : X (F A #> B # m) -> ~ alphabet C a -> bi_tab (X,C) ((X cup (sg (T A # l a) cup sg (F B # l a)), C cup sg (m ~~ l a))::nil).
  Proof. solve. Qed.

  Proposition bi_tableau_exp_TI m : X (T I # m) -> bi_tab (X,C) ((X cup empty, C cup sg (eps ~~ m))::nil).
  Proof. solve. Qed.

  Proposition bi_tableau_exp_Tmult A B a b m : X (T A * B # m) -> a<> b -> ~ alphabet C a -> ~ alphabet C b 
                                             -> bi_tab (X,C) ((X cup (sg (T A # l a) cup sg (T B # l b)), C cup sg (l a o l b ~~ m))::nil).
  Proof. solve. Qed.

  Proposition bi_tableau_exp_Fmult A B x y m : X (F A * B # m) -> pmo C (x o y ~~ m)
                                             -> bi_tab (X,C) ((X cup sg (F A # x), C cup empty)::
                                                              (X cup sg (F B # y), C cup empty)::nil).
  Proof. solve. Qed.

  Proposition bi_tableau_exp_Twand A B x y m : X (T A -* B # m) -> pmo C (x o m ~~ y)
                                            -> bi_tab (X,C) ((X cup sg (F A # x), C cup empty)::
                                                             (X cup sg (T B # y), C cup empty)::nil).
  Proof. solve. Qed.

  Proposition bi_tableau_exp_Fwand A B a b m : X (F A -* B # m) -> a<> b -> ~ alphabet C a -> ~ alphabet C b 
                                             -> bi_tab (X,C) ((X cup (sg (T A # l a) cup sg (F B # l b)), C cup sg (l a o m ~~ l b))::nil).
  Proof. solve. Qed.

End BI_Tableaux_basic.

Section BI_Tableaux_props_1.

  Proposition bi_tab_CSS_inc css1 css2 t : CSS_inc css1 css2 
                                         -> bi_tab css1 t 
                                         -> exists t', bi_tab css2 t' 
                                                    /\ listbranch_eq CSS_inc t t'.
  Proof.
    intros G1 G2.
    apply tableau_prop_inc with (4 := G2); auto.

    intros (X1,C1) (X2,C2) (X3,C3) (X4,C4) [ H1 H2 ] [ H3 H4 ] (H5 & H6 & H7).
    split.
    apply id_transitive with (2 := H3), id_transitive with (2 := H5), id_symetric; auto.
    split.
    apply pmo_id in H2.
    apply pmo_id in H4.
    apply inc_transitive with (2 := proj1 H4),
          inc_transitive with (2 := H6), H2.
    apply alphabet_id in H2.
    apply alphabet_id in H4.
    apply id_transitive with (2 := H4), id_transitive with (2 := H7), id_symetric; auto.

    intros (X,C) (X',C') lb  (H1 & H2 & H3) [ r p Hr Hp ].
    exists (expand (X',C') (bi_inst r p)).
    split; auto.
    apply bi_fireable_inc with (1 := H1) (2 := H2) (4 := Hr) in Hp; auto.
    constructor 1 with (2 := Hp); auto.
    apply listbranch_eq_expand_CSS_inc; auto.
    repeat (split; auto).
  Qed.

End BI_Tableaux_props_1.

Section BI_Tableaux_props_2.

  Variable sigma : label -> label.
  Hypothesis sigma_inj : injective sigma.
    
  Let s a := l (sigma a).

  Proposition bi_tab_subst css t : bi_tab css t -> bi_tab (subst_CSS s css) (map (subst_CSS s) t).
  Proof.
    generalize CSS_id_trans; intros ?.
    apply tableau_prop_fun; auto.
    apply bi_expansion_subst_inj; auto.
  Qed.

End BI_Tableaux_props_2.

Section BI_Tableaux_props_3.

  Variables Xe : stm -> Prop.
  Variables Ce : cst -> Prop.

  Ltac it1 x := apply inc_transitive with (1 := x).
  Ltac it2 x := apply inc_transitive with (2 := x).

  Ltac reduce := repeat
    match goal with 
        [ |- tab_constraints ( _ ++ _ ) inc _ ] => apply inc_transitive with (1 := proj1 (tab_constraints_app _ _))
      | [ |- _ inc tab_constraints ( _ ++ _ ) ] => apply inc_transitive with (2 := proj2 (tab_constraints_app _ _))
      | [ |- tab_constraints ( _ :: _ ) inc _ ] => simpl
      | [ |- _ inc tab_constraints ( _ :: _ ) ] => simpl
      | [ |- ?x cup _ inc ?x cup _ ] => apply bin_union_monotone; auto
      | [ |- ?x cap _ inc ?x cap _ ] => apply bin_inter_monotone; auto
      | [ |- _ cup ?x inc _ cup ?x ] => apply bin_union_monotone; auto
      | [ |- _ cap ?x inc _ cap ?x ] => apply bin_inter_monotone; auto
    end.

  Proposition bi_tab_cup : forall X0 C0 t, bi_tab (X0,C0) t 
                                        -> alphabet (tab_constraints t) cap alphabet Ce inc alphabet C0
                                        -> bi_tab (Xe cup X0, Ce cup C0) (expand (Xe,Ce) t).
  Proof.
    generalize CSS_id_trans; intros Htrans.
    intros X0 C0;
    induction 1 as [ (X1,C1) [ Hb1 Hb2 ] | l1 (X1,C1) l2 lb l' H1 H2 H3 H4 ]; intros Halpha.
    
    simpl expand; unfold expand_1.
    apply tableau_eq with (b0 := (Xe cup X0, Ce cup C0))
                          (t0 := (Xe cup X0, Ce cup C0)::nil); auto.
    simpl; split; auto; simpl; split; auto.
    apply tableau_start; auto.
    
    match goal with [ _: ?x -> bi_tab _ _ |- _ ] => assert x as H5 end.
      it2 Halpha. 
      reduce. 
      apply tab_constraints_id, alphabet_id in H2.
      it2 (proj1 H2).
      apply alphabet_inc.
      reduce.
      apply bi_expansion_increase with (1 := H1).   
    generalize (H4 H5); clear H4; intros H4.

    apply tableau_eq with (b0 := (Xe cup X0, Ce cup C0))
                          (t0 := (expand (Xe,Ce) l1)++(expand (Xe,Ce) lb)++(expand (Xe,Ce) l2)); auto.
    apply expand_id with (css1 := (Xe,Ce)) (css2 := (Xe,Ce)) in H2; auto.
    apply listbranch_eq_tran with (3 := H2); auto.
    do 2 rewrite expand_app; auto.

    rewrite expand_app in H4; simpl in H4.
    destruct bi_expansion_cup with (Xe := Xe) (Ce := Ce) (1 := H1) as ( lb' & H6 & H7 ).

    destruct (bi_tab_increase _ _ H3 (X1,C1)) as [ _ HC1 ].
    apply carrier_app; pred_solve.
    simpl in HC1; apply alphabet_inc in HC1.
    it2 HC1.
    it2 Halpha.
    reduce.
    apply alphabet_inc; auto.
    apply tab_constraints_id in H2.
    it2 (proj1 H2).
    reduce.
    apply inc_transitive with (2 := bin_union_inc_r2 _ _ _).
    reduce; auto.
    
    unfold expand_1 in H4.
    apply tableau_next with (2 := H6) in H4; auto.
    apply tableau_eq with (6 := H4); auto.
  Qed.

End BI_Tableaux_props_3.

Section BI_closed_Tableaux_mono.

  Variables X0 X1 : stm -> Prop.
  Variables C0 C1 : cst -> Prop.

  Hypothesis H0 : finite_CSS (X0,C0).
  Hypothesis H1 : finite_CSS (X1,C1).
  Hypothesis H21 : X0 inc X1.
  Hypothesis H22 : C0 inc C1.

  Proposition bi_has_closed_tableau_monotonic : bi_has_closed_tableau (X0,C0) -> bi_has_closed_tableau (X1,C1).
  Proof.
    generalize CSS_id_trans; intros Htrans.
    destruct H0 as ( H01 & H02 & H03 ).
    destruct H1 as ( H11 & H12 & H13 ).
    destruct (label_hole_injection (alphabet C0) (alphabet C1)) as (sigma & Hs1 & Hs2); auto.
    assert (forall x, alphabet C0 x -> (fun a => l (sigma a)) x = (fun a => l a) x) as Hs3.
      intros x Hx. f_equal; destruct (Hs2 x); tauto.
    assert (forall m, carrier m inc alphabet C0 -> subst_w (fun a => l (sigma a)) m = m) as Hs4.
      intros m Hm.
      rewrite -> (subst_carrier_eq _ _ Hs3 Hm). 
      apply subst_w_id.
    assert (forall x, alphabet C1 (sigma x) -> alphabet C0 (sigma x)) as Hs5.
      intros x HX.
      destruct (Hs2 x) as [ [ G1 G2 ] | [ G1 G2 ] ].
      rewrite G2; auto.
      contradict G2; auto.

    (* we have a substitution of letters that 
          - is injective
          - will rename any letter in C1\C0 into some letter outside C1 
          - is stable on the letters of C0
     *)

    intros [ t [ Ht1 Ht2 ] ].
    generalize Ht1; intros Ht3.
    apply bi_tab_subst with (1 := Hs1) in Ht3.
    assert (CSS_id (subst_CSS (fun a => l (sigma a)) (X0,C0)) (X0,C0)) as Ht5. 
      apply CSS_id_trans with (2 := subst_CSS_id' _).
      apply subst_CSS_alpha_id; auto.
    set (st := map (subst_CSS (fun a => l sigma a)) t); fold st in Ht3.
    apply tableau_eq with (b1 := (X0,C0)) (t1 := st) in Ht3; auto.
    assert (alphabet (tab_constraints st) cap alphabet C1 inc alphabet C0) as Ht7.
      unfold st.
      apply bi_tab_constraints_subst.
      intros x y Hy; apply bin_inter_sg in Hy.
      destruct Hy as [ [ Hy1 | [] ] Hy2 ].
      rewrite <- Hy1 in Hy2 |- *; auto.
    apply bi_tab_cup with (Xe := X1) (Ce := C1) in Ht3; auto.
    exists (expand (X1,C1) st).
    split.
    apply tableau_eq with (6 := Ht3); auto.
    apply bi_cb_expand_carrier; auto.
    apply bi_cb_subst_carrier; auto.
  Qed.

End BI_closed_Tableaux_mono.

Section BI_closed_tableau_pmo_monotonic.

  Variables X0 X1 : stm -> Prop.
  Variables C0 C1 : cst -> Prop.

  Hypothesis H0 : finite_CSS (X0,C0).
  Hypothesis H1 : finite_CSS (X1,C1).
  Hypothesis H21 : X0 inc X1.
  Hypothesis H22 : C0 inc pmo C1.

  Proposition bi_has_closed_tableau_pmo_monotonic : bi_has_closed_tableau (X0,C0) -> bi_has_closed_tableau (X1,C1).
  Proof.
    intros H.
    apply bi_has_closed_tableau_monotonic with (X1 := X1) (C1 := C0 cup C1) in H; auto.
    destruct H as ( t & Ht1 & Ht2 ).
    apply bi_tab_CSS_inc with (css2 := (X1,C1)) in Ht1.
    destruct Ht1 as ( t' & Ht3 & Ht4 ).
    exists t'. 
    split; auto.
    apply bi_cb_inc_carrier with (2 := Ht2); auto.
    split; auto.
    split; auto.
    split; auto.
    apply inc_transitive with (1 := proj1 (alphabet_pmo _)),
          inc_transitive with (2 := proj2 (alphabet_pmo _)),
          alphabet_inc; auto.
    destruct H0 as [ ? [] ].
    destruct H1 as [ ? [] ].
    auto.
  Qed.

End BI_closed_tableau_pmo_monotonic.
 
Corollary bi_has_closed_tableau_eps X C : finite_CSS (X,C) -> bi_has_closed_tableau (X,C cup sg (eps~~eps)) -> bi_has_closed_tableau (X,C).
Proof.
  intros H; generalize H; intros [ ? [ ? ? ] ].
  apply bi_has_closed_tableau_pmo_monotonic; auto.
Qed.

Section bi_proof_subst.

  Variable so : label -> label.
  Variable X : stm -> Prop.
  Variable C : cst -> Prop.

  Hypothesis HC : forall x, alphabet C x \/ ~ alphabet C x.  
  Hypothesis HXC : CSS (X,C).

  Hypothesis Hso1 : forall a b, ~ alphabet C a -> ~ alphabet C b -> so a = so b -> a = b.
  Hypothesis Hso2 : pr_image so (pd_image so (alphabet C)) inc alphabet C.

  Let Hso1' D : C inc D -> forall a b, ~ alphabet D a -> ~ alphabet D b -> so a = so b -> a = b.
  Proof.
    intros H ? ? Ha Hb; apply Hso1; [ contradict Ha | contradict Hb ]; apply alphabet_inc with (1 := H); auto.
  Qed.

  Let Hso2' D : C inc D -> pr_image so (pd_image so (alphabet D)) inc alphabet D.
  Proof.
    intros H x (y & ( z & H1 & H2) & H3).
    destruct (HC z) as [ | H4 ].
    apply alphabet_inc with (1 := H).
    apply Hso2.
    exists y; split; auto.
    exists z; split; auto.
    destruct (HC x) as [ | H5 ].
    apply alphabet_inc with (1 := H); auto.
    rewrite (Hso1 _ _ H5 H4); auto.
    rewrite H2; auto.
  Qed.

  Let s x := l (so x).
  
  Proposition bi_proof_subst t :  bi_tab (X, C) t ->  bi_tab (subst_CSS s (X,C)) (map (subst_CSS s) t).
  Proof.
    generalize CSS_id_trans; intros Htrans.
    induction 1 as [ (X1,C1) [ Hb1 Hb2 ] | l1 (X1,C1) l2 lb l' H1 H2 H3 H4 ].
    
    simpl.
    constructor 1.
    split.
    unfold subst_pstm; auto.
    unfold subst_dimage; auto.
    
    rewrite map_app in H4; simpl in H4.
    apply listbranch_eq_map with (f := subst_CSS s) in H2; auto.
    destruct (bi_tab_increase _ _ H3 (X1,C1)) as [ _ H5 ].
    apply carrier_app; pred_solve.
    destruct (bi_expansion_subst_stable) with (sigma := so) (3 := H1) as ( lb' & H7 & H8 ); simpl;
      try (apply alphabet_inc with (1 := H5); auto).
    apply Hso1'; auto.
    apply Hso2'; auto.    
    apply tableau_eq with (b0 := subst_CSS s (X,C))
                          (t0 := map (subst_CSS s) l1 ++ lb' ++ map (subst_CSS s) l2); auto.
    apply listbranch_eq_tran with (3 := H2); auto.
    do 2 rewrite map_app; auto.
    apply tableau_next with (3 := H4); auto.
  Qed.

End bi_proof_subst.

Section BI_proofs.

  Variable a b : label.
  Variable A : formula.

  Proposition bi_has_proof_ab : bi_has_closed_tableau (sg (F A # l a), sg (l a ~~ l a)) -> bi_has_closed_tableau (sg (F A # l a), sg (l a ~~ l b)).
  Proof.
    apply bi_has_closed_tableau_pmo_monotonic; auto; 
    [  split; auto; apply CSS_sg_l | pred_solve ]; 
    apply pmo_lft with (y := l b), pmo_inc; auto.
  Qed.

  Proposition bi_has_proof_aa : bi_has_closed_tableau (sg (F A # l a), sg (l a ~~ l b)) -> bi_has_closed_tableau (sg (F A # l a), sg (l a ~~ l a)).
  Proof.
    generalize CSS_id_trans; intros Htrans.
    intros ( t & Ht1 & Ht2 ).
    apply bi_proof_subst with (so := subst_one_letter b a) in Ht1.
    simpl in Ht1.
    exists (map (subst_CSS (fun x => l (subst_one_letter b a x))) t).
    split.
    apply tableau_eq with (6 := Ht1); auto.
    split.

    unfold subst_pstm.
    apply id_transitive with (1 := pd_image_sg _ _ _ _).
    simpl; rewrite subst_one_letter_prop_2; auto.

    unfold subst_dimage.
    apply id_transitive with (1 := pd_image_sg _ _ _ _).
    simpl; rewrite subst_one_letter_prop_0, subst_one_letter_prop_2; auto.

    apply bi_cb_subst_carrier; auto.

    apply finite_in_dec; auto.

    intros x y Hx Hy.
    rewrite subst_one_letter_prop_1, subst_one_letter_prop_1; auto.
    contradict Hy; rewrite <- Hy; apply alphabet_sg; pred_solve.
    contradict Hx; rewrite <- Hx; apply alphabet_sg; pred_solve.

    intros x (y & (z & H1 & H2) & H3).
    apply alphabet_sg, bin_union_sg in H1.
    apply alphabet_sg.
    destruct (label_eq_dec b x) as [ Hx | ].
    rewrite <- Hx; pred_solve.
    rewrite subst_one_letter_prop_1 in H3; auto.
    rewrite H3; clear dependent x.
    rewrite <- H2; clear H2.
    destruct H1 as [ [ H1 | [] ] | [ H1 | [] ] ].
    rewrite <- H1, subst_one_letter_prop_2; pred_solve.
    rewrite <- H1, subst_one_letter_prop_0; pred_solve.
  Qed.

End BI_proofs.
 


