(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Classical.

Section Classical_help.

  Variable   A : Prop.
  Let Adec : A \/ ~ A.
  Proof. apply classic. Qed.

  Proposition contraposition B : (~ A -> ~ B) -> B -> A.
  Proof.
    intros H HB.
    destruct Adec as [ | HA ]; trivial.
    contradiction (H HA HB).
  Qed.

  Variable U V : Type.
  Variable P   : U -> V -> Prop.

  Proposition not_all_all :  ~ (forall x y, P x y) -> exists x, exists y, ~ P x y.
  Proof.
    intros H. apply not_all_not_ex. intros H'. apply H. intros x y. apply (not_ex_not_all _ _ (H' x)).
  Qed.

End Classical_help.

Proposition not_all_imp_1 {U} (P Q : U -> Prop) : ~ (forall x, P x -> Q x) -> exists x, P x /\ ~ Q x.
Proof.
  intros H.
  apply not_all_ex_not in H.
  destruct H as ( x & H ).
  exists x.
  apply NNPP; tauto.
Qed.

Proposition not_all_imp_2 {U V} (P Q : U -> V -> Prop) : ~ (forall x y, Q x y -> P x y) -> exists x, exists y, Q x y /\ ~ P x y.
Proof.
  intros H.
  apply not_all_all in H.
  destruct H as ( x & y & H ). 
  apply imply_to_and in H.
  exists x; exists y; auto.
Qed.

Proposition not_all_imp_imp_2 {U V} (P Q R : U -> V -> Prop) : ~ (forall x y, P x y -> Q x y -> R x y) 
	                                                       -> exists x, exists y, P x y /\ Q x y /\ ~ R x y.
Proof.
  intros H.
  apply not_all_all in H.
  destruct H as ( x & y & H ).
  exists x; exists y.
  apply imply_to_and in H.
  destruct H as [ H1 H2 ].
  apply imply_to_and in H2.
  tauto.
Qed.

Proposition not_all_imp_imp_1 {U} (P Q R : U -> Prop) : ~ (forall x, P x -> Q x -> R x) -> exists x, P x /\ Q x /\ ~ R x.
Proof.
  intros H.
  apply not_all_ex_not in H.
  destruct H as [ x H ].
  exists x.
  apply imply_to_and in H.
  destruct H as [ H1 H2 ].
  apply imply_to_and in H2.
  tauto.
Qed.

(* this next one is provable without classical reasonning *)
  
Proposition not_ex_ex U V (P : U -> V -> Prop) : ~ (exists x, exists y, P x y) -> forall x y, ~ P x y.
Proof.
  intros H x y Hxy.
  apply H. exists x. exists y. trivial.
Qed.

Proposition not_or (A B : Prop) : ~ (A \/ B) -> ~ A /\ ~ B.
Proof. tauto. Qed.

Proposition not_not_and (A B : Prop) : ~ (~ A /\ ~ B) -> A \/ B.
Proof.
  intros H.
  apply not_and_or in H.
  destruct H; [ left | right ]; apply NNPP; auto.
Qed.

Ltac classic_solve := 
  repeat match goal with
    | [ H: ~ (_ -> False)              |- _ ] => apply NNPP        in H
    | [ H: ~ (~ _)                     |- _ ] => apply NNPP        in H
    | [ H: ~ (~ _ /\ ~ _)              |- _ ] => apply not_not_and in H	
    | [ H: ~ (_ \/ _)                  |- _ ] => apply not_or      in H 
    | [ H: ~ (_ /\ _)                  |- _ ] => apply not_and_or  in H
    | [ H: ~ (forall _, _ -> _ -> _)   |- _ ] => apply not_all_imp_imp_1  in H
    | [ H: ~ (forall _, _ -> _)        |- _ ] => apply not_all_imp_1 in H
    | [ H: ~ (forall _ _, _ -> _ -> _) |- _ ] => apply not_all_imp_imp_2  in H
    | [ H: ~ (forall _ _, _ -> _)      |- _ ] => apply not_all_imp_2 in H
    | [ H: ~ (forall _ _, _)           |- _ ] => apply not_all_all in H
    | [ H: _ -> _                      |- _ ] => apply imply_to_or in H
  end.

Ltac classic_disj x y := apply not_not_and; intros [ x y ].