(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import Omega.

Require Import predicate.
Require Import carrier.
Require Import finite.
Require Import maps.

Section Nat_hole.

  (* if A and B are two finite subsets of nat with A included in B
     then there exists a function which is the identity on A and 
     is outside B for every element not in A 
  *)

  Variables A B  : nat -> Prop.
  Hypothesis HA  : finite A.
  Hypothesis HB  : finite B.
  Hypothesis HAB : A inc B.

  Proposition nat_hole_injection : exists f, injective f 
                                         /\  forall x, (  A x /\ f x = x) 
                                                    \/ (~ A x /\ ~ B (f x)).
  Proof.
    destruct (finite_nat_bounded _ HB) as [ m Hm ].
    destruct (finite_bool_dec _ eq_nat_dec _ HA) as ( dA & Hd1 & Hd2 ).
    set (f x := match dA x with true => x | false => S (m+x) end).
    exists f.
    split.

    intros x y Hxy; unfold f in Hxy.
    case_eq (dA x); case_eq (dA y); 
    intros Hx Hy; rewrite Hx, Hy in Hxy; auto.
    apply Hd2, HAB, Hm in Hy; omega.
    apply Hd2, HAB, Hm in Hx; omega.
    omega.

    intros x. unfold f.
    case_eq (dA x); intros Hx.
    apply Hd2 in Hx. left; auto.
    right.
    split; intros H. apply Hd1 in H.
    rewrite Hx in H; discriminate H.
    apply Hm in H. omega.
  Qed.

End Nat_hole.


  
