(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Proposition replace_type_eq (G H : Type) : G = H -> G -> H.
Proof. intros H1 ?; rewrite <- H1; auto. Qed.

Proposition replace_set_eq (G H : Set) : G = H -> G -> H.
Proof. intros H1 ?; rewrite <- H1; auto. Qed.

Proposition replace_prop_eq (G H : Prop) : G = H -> G -> H.
Proof. intros H1 ?; rewrite <- H1; auto. Qed.

Ltac replace_prop_with K := apply replace_prop_eq with (2 := K); repeat (f_equal; auto).
Ltac replace_set_with K := apply replace_set_eq with (2 := K); repeat (f_equal; auto).
Ltac replace_type_with K := apply replace_type_eq with (2 := K); repeat (f_equal; auto).

Ltac replace_with K := 
  let H := fresh
  in  generalize K; intros H;
  match goal with [ H: ?x |- _ ] => 
    match type of x with
      | Prop => replace_prop_with H
      | Set  => replace_set_with H
      | Type => replace_type_with H
    end
  end; clear H.

Section Replace_example.

  Variable P : nat -> Prop.

  Let test x y : x = y -> P x -> P y.
  Proof. intros ? H2; replace_with H2. Qed.

End Replace_example.

