(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import predicate.
Require Import carrier.

Require Import labels.
Require Import words.

Definition subst_one_letter (a b : label) x := match label_eq_dec a x with left _ => b | right _ => x end.

Proposition subst_one_letter_prop_0 a b : subst_one_letter a b a = b.
Proof. unfold subst_one_letter; destruct (label_eq_dec a a) as [ _ | H ]; auto; contradiction H; auto. Qed. 

Proposition subst_one_letter_prop_1 a b x : a <> x -> subst_one_letter a b x = x.
Proof. unfold subst_one_letter; destruct (label_eq_dec a x) as [ G | _ ]; auto; intros H; contradict H; auto. Qed. 

Proposition subst_one_letter_prop_2 a b : subst_one_letter a b b = b.
Proof. unfold subst_one_letter; destruct (label_eq_dec a b) as [ G | _ ]; auto; intros H; contradict H; auto. Qed. 

Definition subst_l (sigma : label -> label) := subst_w (fun (x : label) => l x).

Definition subst_comp (s s' : label -> word label) x := subst_w s' (s x).

Definition subst_one_l (a : label) (w : word _) := fun x => match label_eq_dec x a with left _ => w | right _ => l x end.
Definition subst_one_w a w m := subst_w (subst_one_l a w) m.

Proposition subst_one_l_prop_0 a w : subst_one_l a w a = w.
Proof. simpl. unfold subst_one_l. destruct (label_eq_dec a a) as [ _ | H ]. reflexivity. contradiction H. reflexivity. Qed. 

Proposition subst_one_w_prop_0 a w : subst_one_w a w (l a) = w.
Proof. simpl. rewrite -> subst_one_l_prop_0. symmetry. apply app_nil_end. Qed. 

Proposition subst_one_w_prop_2 a w m : ~ (carrier m) a -> subst_one_w a w m = m.
Proof.
  induction m as [ | b m ]; intros H; simpl; auto.
  unfold subst_one_l at 1. 
  destruct (label_eq_dec b a) as [ Hab | Hab ].
  contradiction H; rewrite -> Hab; auto.
  rewrite -> IHm; auto.
Qed.

Section change_two.

  Set Implicit Arguments.

  Variable M : Type.

  Definition change_two (rho : label -> word M) a x b y c := 
    match (label_eq_dec c a, label_eq_dec c b) with
      | (left _, _) => x
      | (_, left _) => y
      | _              => rho c
    end.
    
  Proposition change_two_prop_a rho a x b y : change_two rho a x b y a = x.
  Proof.
    unfold change_two.
    destruct (label_eq_dec a a) as [ _ | H ]; auto.
    contradiction H; auto.
  Qed.
  
  Proposition change_two_prop_b rho a x b y : a<>b -> change_two rho a x b y b = y.
  Proof.
    intros H0; unfold change_two.
    destruct (label_eq_dec b a) as [ H | _ ]. symmetry in H. contradiction (H0 H).
    destruct (label_eq_dec b b) as [ _ | H ]. reflexivity. contradiction H. reflexivity.
  Qed.
  
  Proposition change_two_prop_ab rho a x b y c : c <> a -> c <> b -> change_two rho a x b y c = rho c.
  Proof.
    intros Ha Hb. unfold change_two.
    destruct (label_eq_dec c a) as [ H | _ ]. contradiction (Ha H).
    destruct (label_eq_dec c b) as [ H | _ ]. contradiction (Hb H).
    reflexivity.
  Qed.
  
  Definition change_two_l (rho : label -> M) a x b y c := 
    match (label_eq_dec c a, label_eq_dec c b) with
      | (left _, _) => x
      | (_, left _) => y
      | _              => rho c
    end.

  Proposition change_two_l_prop_a rho a x b y : change_two_l rho a x b y a = x.
  Proof.
    unfold change_two_l.
    destruct (label_eq_dec a a) as [ _ | H ].
    reflexivity. contradiction H. reflexivity.
  Qed.
  
  Proposition change_two_l_prop_b rho a x b y : a <> b -> change_two_l rho a x b y b = y.
  Proof.
    intros H0. unfold change_two_l.
    destruct (label_eq_dec b a) as [ H | _ ]. symmetry in H. contradiction (H0 H).
    destruct (label_eq_dec b b) as [ _ | H ]. reflexivity. contradiction H. reflexivity.
  Qed.
  
  Proposition change_two_l_prop_ab rho a x b y c : c <> a -> c <> b -> change_two_l rho a x b y c = rho c.
  Proof.
    intros Ha Hb. unfold change_two_l.
    destruct (label_eq_dec c a) as [ H | _ ]. contradiction (Ha H).
    destruct (label_eq_dec c b) as [ H | _ ]. contradiction (Hb H).
    reflexivity.
  Qed.
  
  Proposition change_two_prop_carrier rho a x b y k : ~ carrier k a -> ~ carrier k b -> subst_w (change_two rho a x b y) k = subst_w rho k.
  Proof.
    intros Ha Hb.
    apply subst_carrier_eq with (P := carrier k).
    intros c Hc. apply change_two_prop_ab; intros H ; [ apply Ha | apply Hb ]; rewrite <- H; trivial.
    apply inc_reflexive.
  Qed.
  
  Proposition subst_commute s s' m : subst_w s' (subst_w s m) = subst_w (subst_comp s s') m.
  Proof.
    induction m.
    simpl. reflexivity.
    simpl. rewrite -> subst_o.
    rewrite -> IHm. unfold subst_comp. reflexivity.
  Qed.

End change_two.


